// Window class implementation
#include "pch.h"

#include "Window.h"
#include "util.h"

namespace nf::client {
    Window::Window(const char* windowTitle)
        : m_handle(nullptr)
        , m_wndClassName("NothinFancyWindow")
        , m_styleWindowed(WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX)
        , m_styleFullscreen(WS_POPUP)
        , m_currentWidth()
        , m_currentHeight()
    {
        NFLog("Creating window");

        // Disable automatic DPI upscaling
        SetProcessDpiAwarenessContext(DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2);

        registerWindowClass();

        m_handle = CreateWindow(m_wndClassName, windowTitle, NULL, 0, 0, 0, 0, nullptr, nullptr, nullptr, this);
    }

    HWND Window::getHandle() const {
        return m_handle;
    }

    void Window::setDisplay(DisplayConfig& config) {
        NFLog("Setting window display");

        // TODO: Only use "active" monitor when starting windowed
        POINT cursor = {};
        GetCursorPos(&cursor);
        MONITORINFO mi = {};
        mi.cbSize = sizeof(mi);
        GetMonitorInfo(MonitorFromPoint(cursor, MONITOR_DEFAULTTONEAREST), &mi);
        int monitorX = mi.rcMonitor.left, monitorY = mi.rcMonitor.top;
        int monitorWidth = mi.rcMonitor.right - monitorX, monitorHeight = mi.rcMonitor.bottom - monitorY;
        int windowX = 0, windowY = 0;
        unsigned int windowWidth = 0, windowHeight = 0;

        switch (config.mode) {
            case DisplayMode::Windowed: {
                SetWindowLongPtr(m_handle, GWL_STYLE, m_styleWindowed);
                m_currentWidth = config.width, m_currentHeight = config.height;
                windowX = monitorX + (monitorWidth / 2) - (m_currentWidth / 2), windowY = monitorY + (monitorHeight / 2) - (m_currentHeight / 2);
                SIZE windowSize = getWindowSize();
                windowWidth = windowSize.cx, windowHeight = windowSize.cy;
                break;
            }

            case DisplayMode::Fullscreen: {
                SetWindowLongPtr(m_handle, GWL_STYLE, m_styleFullscreen);
                windowX = monitorX, windowY = monitorY;
                windowWidth = monitorWidth, windowHeight = monitorHeight;
                config.width = windowWidth, config.height = windowHeight;
                break;
            }
        }

        SetWindowPos(m_handle, nullptr, windowX, windowY, windowWidth, windowHeight, SWP_NOZORDER | SWP_FRAMECHANGED);
    }

    void Window::runLoop() {
        // At some point, the window started to show unselected at first.
        SetFocus(m_handle);
        MSG msg = {};
        while (GetMessage(&msg, nullptr, NULL, NULL)) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }

    void Window::show(bool show) {
        ShowWindow(m_handle, show ? SW_SHOW : SW_HIDE);
    }

    void Window::registerWindowClass() {
        WNDCLASS wc = {};
        wc.lpszClassName = m_wndClassName;
        wc.hCursor = LoadCursor(nullptr, IDC_ARROW);
        wc.lpfnWndProc = wndProc;
        RegisterClass(&wc);
    }

    LRESULT CALLBACK Window::wndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
        static Window* wnd = nullptr;

        switch (msg) {
            case WM_CREATE:
                wnd = reinterpret_cast<Window*>(reinterpret_cast<CREATESTRUCT*>(lParam)->lpCreateParams);
                return 0;

            case WM_DPICHANGED: {
                // Prevents automatic window resize on DPI change (don't apply to fullscreen)
                if (GetWindowLongPtr(hWnd, GWL_STYLE) == wnd->m_styleFullscreen)
                    return 0;

                SIZE windowSize = wnd->getWindowSize();
                SetWindowPos(hWnd, nullptr, 0, 0, windowSize.cx, windowSize.cy, SWP_NOZORDER | SWP_NOMOVE);
                return 0;
            }

            case WM_CLOSE:
                PostQuitMessage(0);
                return 0;
        }

        return DefWindowProc(hWnd, msg, wParam, lParam);
    }

    SIZE Window::getWindowSize() {
        RECT cli = {};
        cli.right = m_currentWidth;
        cli.bottom = m_currentHeight;
        AdjustWindowRectExForDpi(&cli, m_styleWindowed, FALSE, NULL, GetDpiForWindow(m_handle));
        int width = cli.right - cli.left, height = cli.bottom - cli.top;
        return { width, height };
    }

    Window::~Window() {
        DestroyWindow(m_handle);
        NFLog("Window destroyed");
    }
}
